package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class contains code that converts a context free grammar into
 * a DFA for a regular approximation of the language.
 *
 * The basic translation is due to Nederhof and is decribed in the paper
 * "Practical Experiments with Regular pproximation" in Computational
 * Linguistics, 26(1).  Section 4.1.
 */
class ApproximationFA
{
    List terminals;
    List nonterminals;
    List allStates;
    NFAState startState;
    NFAState acceptState;
    GLRNonterminal startSymbol;
    List alphabet;
    Partition partitions;

    /** Map from nonterminals to terminals created for the nonterminal. */
    Label[] ntTerminals;

    /** NFAs for each nonterminal. */
    NFA[] nfas;

    /**
     * Build a regular approximation for the rule's RHS.
     * The LHS must be a augmented grammar start symbol; it cannot
     * be reachable from the RHS.
     */
    ApproximationFA(List terminals, List nonterminals, GLRNonterminal S,
                    Partition partitions)
    {
        this.terminals = new ArrayList();
        this.nonterminals = nonterminals;
        this.startSymbol = S;
        this.partitions = partitions;

        this.allStates = new ArrayList();

        if (S.rules().size() != 1) {
            throw new InternalCompilerError(
                "Cannot create NFA for non-augmented grammar.");
        }

        this.ntTerminals = new Label[nonterminals.size()];
        this.alphabet = new ArrayList(terminals.size() + nonterminals.size());
        alphabet.addAll(terminals);

        // Create a nonterminal-terminal for each nonterminal.
        for (Iterator i = nonterminals.iterator(); i.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) i.next();
            ntTerminal(A);
        }

        // Now for each mutually recursive partition, generate a DFA.
        // Nonterminals not in the partition are treated as terminals.

        this.nfas = new NFA[nonterminals.size()];

        if (partitions.recursive(S) != Partition.NONE) {
            throw new InternalCompilerError("Cannot have recursive start symbol.");
        }

        NonterminalNFA nfa = new NonterminalNFA(S, alphabet());
        nfas[S.index()] = nfa;

        for (Iterator j = S.rules.iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();
            makeFA(nfa, nfa.q0, rule.rhs, nfa.q1);
            nfas[S.index()] = new DFA(nfa);
        }

        for (Iterator k = nonterminals.iterator(); k.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) k.next();

            if (nfas[A.index()] == null &&
                partitions.recursive(A) == Partition.NONE) {

                System.out.println("Missed NFA for non-recursive nonterminal " + A);

                NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                nfas[A.index()] = Anfa;

                for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                    GLRRule rule = (GLRRule) j.next();
                    makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                }

                nfas[A.index()] = new DFA(Anfa);
            }
        }

        for (Iterator k = nonterminals.iterator(); k.hasNext(); ) {
            GLRNonterminal A = (GLRNonterminal) k.next();
            nfas[A.index()].dump();
        }
    }

    public List alphabet() {
        return alphabet;
    }

    class NonterminalNFA extends AbstractNFA {
        NFAState q0, q1;
        GLRNonterminal A;
        List states;

        NonterminalNFA(GLRNonterminal A, List terminals) {
            super(terminals);
            this.states = new ArrayList();
            this.A = A;
            this.q0 = freshState(A.toString() + "0");
            this.q1 = freshState(A.toString() + "1");
        }

        public NFAState start() {
            return q0;
        }

        NFAState freshState(final String s) {
            NFAState q = new NFAState(states.size()) {
                public String toString() { return s; } };
            states.add(q);
            return q;
        }

        NFAState freshState(GLRNonterminal A) {
            NFAState q = new NonterminalState(A, states.size());
            states.add(q);
            return q;
        }

        public BitSet accepting(NFAState q0) {
            BitSet b = new BitSet();
            if (q0 == this.q0) {
                b.set(this.q1.index());
            }
            return b;
        }

        public boolean accepting(NFAState q0, NFAState q1) {
            return q0 == this.q0 && q1 == this.q1;
        }

        public Iterator nodes() {
            return states.iterator();
        }

        public int size() {
            return states.size();
        }

        public NFAState nodeAt(int i) {
            return (NFAState) states.get(i);
        }

        public NFAEdge edge(NFAState q1, Label label, NFAState q2) {
            NFAEdge e = new NFAEdge(q1, q2, label);
            return e;
        }

        public String toString() {
            return "approximation NFA for " + A;
        }
    }

    public NFAState nodeAt(int i) {
        return (NFAState) allStates.get(i);
    }

    public Iterator nodes() {
        return allStates.iterator();
    }

    public NFAState start() { return startState; }
    public NFAState accept() { return acceptState; }

    public BitSet accepting(NFAState q0) {
        BitSet b = new BitSet();
        if (q0 == this.start()) {
            b.set(this.accept().index());
        }
        return b;
    }

    public boolean accepting(NFAState q0, NFAState q) {
        return q0 == start() && q == accept();
    }

    public int size() { return allStates.size(); }

    static class NonterminalState extends NFAState {
        GLRNonterminal A;

        NonterminalState(GLRNonterminal A, int index) {
            super(index);
            this.A = A;
        }

        public String toString() {
            return A.toString();
        }
    }

    static GLRSymbol leftHead(List alpha) {
        if (alpha.size() > 0) {
            return (GLRSymbol) alpha.get(0);
        }

        return null;
    }

    static GLRSymbol rightHead(List alpha) {
        if (alpha.size() > 0) {
            return (GLRSymbol) alpha.get(alpha.size()-1);
        }

        return null;
    }

    static List leftTail(List alpha) {
        if (alpha.size() > 1) {
            return alpha.subList(1,alpha.size());
        }
        return Collections.EMPTY_LIST;
    }

    static List rightTail(List alpha) {
        if (alpha.size() > 1) {
            return alpha.subList(0,alpha.size()-1);
        }
        return Collections.EMPTY_LIST;
    }

    void makeFAsForList(List alpha) {
        System.out.println("makeFAsForList: " + alpha);

        for (Iterator i = alpha.iterator(); i.hasNext(); ) {
            GLRSymbol X = (GLRSymbol) i.next();

            if (X instanceof GLRNonterminal) {
                GLRNonterminal A = (GLRNonterminal) X;

                if (nfas[A.index()] != null) {
                    continue;
                }

                if (partitions.recursive(A) != Partition.SELF) {
                    NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                    nfas[A.index()] = Anfa;
    System.out.println("new " + Anfa);

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                    }

                    // Now create a DFA for the NFA.
                    DFA Adfa = new DFA(Anfa);
                    nfas[A.index()] = Adfa;
                }
                else {
                    Collection p = partitions.members(A);

                    SelfRecursiveApproximationFA pnfa =
                        new SelfRecursiveApproximationFA(alphabet(), nonterminals, this, p);
                    DFA pdfa = new DFA(pnfa, pnfa.startStates());
                    pdfa.lazyBuild();

                    for (Iterator j = p.iterator(); j.hasNext(); ) {
                        GLRNonterminal B = (GLRNonterminal) j.next();
                        NFAState nfaStart = pnfa.startStates[B.index()];
                        DFA.DFAState dfaStart = pdfa.startStatesMap[nfaStart.index()];
                        nfas[B.index()] = new WrapperDFA(pdfa, dfaStart);
                    }

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }
                }
            }
        }
    }

    void makeFA(NonterminalNFA nfa, NFAState q0, List alpha, NFAState q1) {
        System.out.println("makeFA: nfa " + nfa + ", " + alpha);

        if (alpha.isEmpty()) {
            nfa.epsilon(q0, q1);
        }
        else if (alpha.size() > 1) {
            GLRSymbol X = (GLRSymbol) leftHead(alpha);
            NFAState q = nfa.freshState("after-" + X.toString());
            System.out.println("  recurse on " + X);
            makeFA(nfa, q0, Collections.singletonList(X), q);
            System.out.println("  back");
            System.out.println("  recurse on " + leftTail(alpha));
            makeFA(nfa, q, leftTail(alpha), q1);
            System.out.println("  back");
        }
        else if (leftHead(alpha) instanceof GLRTerminal) {
            GLRTerminal a = (GLRTerminal) leftHead(alpha);
            nfa.transition(q0, a, q1);
        }
        else {
            GLRNonterminal A = (GLRNonterminal) leftHead(alpha);

            NFAState[] ntStates = new NFAState[nonterminals.size()];

            switch (partitions.recursive(A)) {
                case Partition.NONE: System.out.println("non-recursive"); break;
                case Partition.LEFT: System.out.println("left"); break;
                case Partition.RIGHT: System.out.println("right"); break;
                case Partition.CYCLIC: System.out.println("cyclic"); break;
                case Partition.SELF: System.out.println("self"); break;
            }

            if (nfas[A.index()] == null) {
                if (partitions.recursive(A) != Partition.SELF) {
                    NonterminalNFA Anfa = new NonterminalNFA(A, alphabet());
                    nfas[A.index()] = Anfa;
    System.out.println("new " + Anfa);

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(Anfa, Anfa.q0, rule.rhs, Anfa.q1);
                    }

                    // Now create a DFA for the NFA.
                    DFA Adfa = new DFA(Anfa);
                    nfas[A.index()] = Adfa;
                }
                else {
                    Collection p = partitions.members(A);

                    SelfRecursiveApproximationFA pnfa =
                        new SelfRecursiveApproximationFA(alphabet(), nonterminals, this, p);
                    DFA pdfa = new DFA(pnfa, pnfa.startStates());
                    pdfa.lazyBuild();

                    for (Iterator j = p.iterator(); j.hasNext(); ) {
                        GLRNonterminal B = (GLRNonterminal) j.next();
                        NFAState nfaStart = pnfa.startStates[B.index()];
                        DFA.DFAState dfaStart = pdfa.startStatesMap[nfaStart.index()];
                        nfas[B.index()] = new WrapperDFA(pdfa, dfaStart);
                    }

                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }
                }
            }

            switch (partitions.recursive(A)) {
                case Partition.LEFT: {
                    if (! partitions.same(A, nfa.A)) {
                        nfa.transition(q0, ntTerminal(A), q1);
                    }
                    else {
                        Collection p = partitions.members(A);

                        for (Iterator i = p.iterator(); i.hasNext(); ) {
                            GLRNonterminal B = (GLRNonterminal) i.next();
                            ntStates[B.index()] = nfa.freshState(B);
                        }

                        NFAState qA = ntStates[A.index()];

                        for (Iterator i = p.iterator(); i.hasNext(); ) {
                            GLRNonterminal C = (GLRNonterminal) i.next();

                            NFAState qC = ntStates[C.index()];

                            for (Iterator j = C.rules.iterator(); j.hasNext(); ) {
                                GLRRule rule = (GLRRule) j.next();

                                if (! partitions.containsAny(A, rule.rhs)) {
                                    makeFA(nfa, q0, rule.rhs, qC);
                                }

                                GLRSymbol D = leftHead(rule.rhs);

                                if (D instanceof GLRNonterminal &&
                                    partitions.contains(A, D) &&
                                    ! partitions.containsAny(A, leftTail(rule.rhs))) {
                                    NFAState qD = ntStates[D.index()];
                                    makeFA(nfa, qD, leftTail(rule.rhs), qC);
                                }
                            }
                        }

                        nfa.epsilon(qA, q1);
                    }

                    break;
                }
                case Partition.RIGHT:
                case Partition.CYCLIC: {
                    if (! partitions.same(A, nfa.A) &&
                        partitions.recursive(A) != Partition.NONE) {
                        nfa.transition(q0, ntTerminal(A), q1);
                    }
                    else {
                        Collection p = partitions.members(A);

                        for (Iterator i = p.iterator(); i.hasNext(); ) {
                            GLRNonterminal B = (GLRNonterminal) i.next();
                            ntStates[B.index()] = nfa.freshState(B);
                        }

                        NFAState qA = ntStates[A.index()];

                        for (Iterator i = p.iterator(); i.hasNext(); ) {
                            GLRNonterminal C = (GLRNonterminal) i.next();

                            NFAState qC = ntStates[C.index()];

                            for (Iterator j = C.rules.iterator(); j.hasNext(); ) {
                                GLRRule rule = (GLRRule) j.next();

                                if (! partitions.containsAny(A, rule.rhs)) {
                                    makeFA(nfa, qC, rule.rhs, q1);
                                }

                                GLRSymbol D = rightHead(rule.rhs);

                                if (D instanceof GLRNonterminal &&
                                    partitions.contains(A, D) &&
                                    ! partitions.containsAny(A, rightTail(rule.rhs))) {
                                    NFAState qD = ntStates[D.index()];
                                    makeFA(nfa, qC, rightTail(rule.rhs), qD);
                                }
                            }
                        }

                        nfa.epsilon(q0, qA);
                    }

                    break;
                }
                case Partition.SELF: {
                    nfa.transition(q0, ntTerminal(A), q1);

                    // Make sure DFAs are build for all the nonterminals
                    // reachable from A.
                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFAsForList(rule.rhs);
                    }

                    break;
                }
                default: {
                    // not recursive
                    for (Iterator j = A.rules.iterator(); j.hasNext(); ) {
                        GLRRule rule = (GLRRule) j.next();
                        makeFA(nfa, q0, rule.rhs, q1);
                    }
                    break;
                }
            }
        }
    }

    static class NTTerminal implements Label {
        GLRNonterminal A;
        int index;
        NTTerminal(GLRNonterminal A, int index) {
            this.A = A;
            this.index = index;
        }
        public int index() { return index; }
        public String toString() {
            return "t" + A.toString();
        }
    }

    Label ntTerminal(GLRNonterminal A) {
        Label a = ntTerminals[A.index()];
        if (a == null) {
            a = new NTTerminal(A, alphabet.size());
            ntTerminals[A.index()] = a;
            alphabet.add(a);
        }
        return a;
    }

    NFA dfaForSymbol(GLRSymbol X) {
        System.out.println("getting DFA for " + X);

        if (X instanceof GLRNonterminal) {
            GLRNonterminal A = (GLRNonterminal) X;
            NFA dfa = nfas[A.index()];
            if (dfa instanceof DFA) {
                ((DFA) dfa).lazyBuild();
            }
            return dfa;
        }
        else {
            return new TerminalDFA((GLRTerminal) X, alphabet);
        }
    }

    static class TerminalDFA implements NFA {
        GLRTerminal label;
        List alphabet;
        NFAState startState;
        NFAState acceptState;

        public TerminalDFA(final GLRTerminal label, List alphabet) {
            this.label = label;
            this.alphabet = alphabet;
            this.startState = new NFAState(0) {
                public String toString() { return label + "0"; } };
            this.acceptState = new NFAState(1) {
                public String toString() { return label + "1"; } };
        }

        public boolean equals(Object o) {
            if (o instanceof TerminalDFA) {
                TerminalDFA m = (TerminalDFA) o;
                return label.equals(m.label);
            }
            return false;
        }

        public String toString() {
            return "TerminalDFA for " + label;
        }

        public int size() {
            return 2;
        }

        public Iterator succs(Node node) {
            if (node == startState) {
                NFAEdge edge = new NFAEdge(startState, acceptState, label);
                return Collections.singleton(edge).iterator();
            }
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator preds(Node node) {
            if (node == acceptState) {
                NFAEdge edge = new NFAEdge(acceptState, startState, label);
                return Collections.singleton(edge).iterator();
            }
            return Collections.EMPTY_LIST.iterator();
        }

        public Iterator nodes() {
            List l = new ArrayList(2);
            l.add(startState);
            l.add(acceptState);
            return l.iterator();
        }

        public NFAState nodeAt(int i) {
            if (i == 0) return startState;
            if (i == 1) return acceptState;
            return null;
        }

        public List alphabet() {
            return alphabet;
        }

        public BitSet accepting(NFAState q0) {
            BitSet b = new BitSet();
            b.set(acceptState.index());
            return b;
        }

        public boolean accepting(NFAState q0, NFAState q) {
            return q0 == startState && q == acceptState;
        }

        public NFAState start() {
            return startState;
        }

        public void dump() {
            System.out.println(toString());
        }
    }

    static class ListNFA implements NFA {
        List nfas;
        List nodes;
        String alpha;
        Map map;

        ListNFA(NFA head, ListNFA tail, String alpha) {
            this.nfas = new ArrayList(1 +
                                      (tail != null ? tail.nfas.size() : 0));
            this.nodes = new ArrayList(head.size() +
                                       (tail != null ? tail.nodes.size() : 0));
            this.map = new HashMap();
            this.alpha = alpha;

            nfas.add(head);

            if (tail != null) {
                nfas.addAll(tail.nfas);
            }

            int nfa_index = 0;

            for (Iterator i = nfas.iterator(); i.hasNext(); ) {
                NFA nfa = (NFA) i.next();
                for (Iterator j = nfa.nodes(); j.hasNext(); ) {
                    NFAState q = (NFAState) j.next();
                    ListNFAState lq = new ListNFAState(nfa_index, q, nodes.size());
                    nodes.add(lq);
                    map.put(new Key(nfa_index, q), lq);
                }

                nfa_index++;
            }
        }
        
        static class Key {
            int nfa_index;
            NFAState q;

            Key(int nfa_index, NFAState q) {
                this.nfa_index = nfa_index;
                this.q = q;
            }

            public int hashCode() {
                return (nfa_index + 3) * q.hashCode();
            }

            public boolean equals(Object o) {
                return o instanceof Key
                    && ((Key) o).nfa_index == nfa_index
                    && ((Key) o).q.equals(q);
            }
        }

        static class ListNFAState extends NFAState {
            int nfa_index;
            NFAState q;

            ListNFAState(int nfa_index, NFAState q, int index) {
                super(index);
                this.nfa_index = nfa_index;
                this.q = q;
            }

            public String toString() {
                return q.toString() + "<" + index + "," + nfa_index + ">";
            }
        }

        public Iterator nodes() {
            return nodes.iterator();
        }

        public int size() {
            return nodes.size();
        }

        private static Key key = new Key(0, null);

        private ListNFAState map(int nfa_index, NFAState q) {
            key.nfa_index = nfa_index;
            key.q = q;
            return (ListNFAState) map.get(key);
        }

        public Iterator succs(Node n) {
            ListNFAState lq = (ListNFAState) n;

            NFA nfa = (NFA) nfas.get(lq.nfa_index);

            List l = new ArrayList();

            if (nfa == null) {
                return Collections.EMPTY_LIST.iterator();
            }

            if (nfa.accepting(nfa.start(), lq.q)) {
                if (lq.nfa_index+1 < nfas.size()) {
                    NFA next = (NFA) nfas.get(lq.nfa_index+1);
                    NFAState q = (NFAState) next.start();
                    l.add(new NFAEdge(lq, map(lq.nfa_index+1, q), null));
                }
            }

            for (Iterator i = nfa.succs(lq.q); i.hasNext(); ) {
                NFAEdge e = (NFAEdge) i.next();
                NFAState q = (NFAState) e.target();
                l.add(new NFAEdge(lq, map(lq.nfa_index, q), e.label()));
            }
            
            return l.iterator();
        }

        public Iterator preds(Node n) {
            throw new InternalCompilerError("unimplemented");
        }

        public NFAState start() {
            NFA nfa0 = (NFA) nfas.get(0);
            return map(0, nfa0.start());
        }

        public BitSet accepting(NFAState q0) {
            ListNFAState lq0 = (ListNFAState) q0;

            if (lq0 != start()) {
                return new BitSet();
            }

            NFA nfa = (NFA) nfas.get(nfas.size()-1);

            BitSet a = nfa.accepting(nfa.start());
            BitSet b = new BitSet();

            for (int i = b.nextSetBit(0); i >= 0; i = b.nextSetBit(i+1)) {
                NFAState q = (NFAState) nfa.nodeAt(i);
                NFAState lq = map(nfas.size()-1, q);
                b.set(lq.index());
            }

            return b;
        }

        public boolean accepting(NFAState q0, NFAState q) {
            ListNFAState lq0 = (ListNFAState) q0;

            if (lq0 != start()) {
                return false;
            }

            ListNFAState lq = (ListNFAState) q;

            NFA nfa = (NFA) nfas.get(lq.nfa_index);
            return nfa.accepting(nfa.start(), lq.q) && lq.nfa_index+1 == nfas.size();
        }

        public NFAState nodeAt(int i) {
            return (NFAState) nodes.get(i);
        }

        public List alphabet() {
            NFA nfa0 = (NFA) nfas.get(0);
            return nfa0.alphabet();
        }

        public String toString() {
            return "list nfa for " + alpha;
        }

        public void dump() {
            if (Report.should_report("approx", 3)) {
                Report.report(1, "Dumping " + this);
                
                System.out.println();
                System.out.println("digraph G {");

                for (Iterator i = nodes(); i.hasNext(); ) {
                    NFAState n = (NFAState) i.next();
                    System.out.println("\"" + n + "\"");
                    for (Iterator j = succs(n); j.hasNext(); ) {
                        NFAEdge e = (NFAEdge) j.next();
                        System.out.println("/* " + e + " */");
                        System.out.print("  \"" + e.source() + "\"");
                        System.out.print("->\"" + e.target() + "\"");
                        if (e.label() != null) {
                            System.out.println(" [ label=\"" + e.label() + "\" ]"); 
                        } 
                        else {
                            System.out.println();
                        }
                    }
                }

                System.out.println("}");
                System.out.println();
            }
        }
    }

    ListNFA forList(List alpha) {
        if (! alpha.isEmpty()) {
            NFA head = dfaForSymbol(leftHead(alpha));
            ListNFA tail = forList(leftTail(alpha));
            return new ListNFA(head, tail, alpha.toString());
        }

        return null;
    }

    /**
     * Return true if the language of symbols <code>alpha1</code> intersects
     * the language of symbols <code>alpha2</code>.
     */
    public boolean intersects(List alpha1, List alpha2) {
        if (alpha1.isEmpty() || alpha2.isEmpty()) {
            return alpha1.isEmpty() && alpha2.isEmpty();
        }

        NFA nfa1 = forList(alpha1);
        NFA nfa2 = forList(alpha2);

        nfa1.dump();
        nfa2.dump();

        return NFAIntersect.intersects(nfa1, nfa2);

        // DFA dfa1 = new DFA(nfa1);
        // DFA dfa2 = new DFA(nfa2);
        // return dfa1.intersects(dfa2);
    }
}
